/**
 * 
 */
package de.mpc.prolic.prolic;

import java.io.File;
import java.util.Comparator;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.ArrayList;
import java.util.Random;
import java.util.TreeMap;
import java.util.TreeSet;
import java.util.Collection;
import java.util.Set;

import org.apache.logging.log4j.Logger; 
import org.apache.logging.log4j.LogManager;
import org.biojava.nbio.core.exceptions.CompoundNotFoundException;
import org.jgrapht.UndirectedGraph;
import org.jgrapht.alg.BronKerboschCliqueFinder;
import org.jgrapht.graph.DefaultEdge;
import org.jgrapht.graph.SimpleGraph;

import de.mpc.prolic.comparison.PeptideBasedComparison;
import de.mpc.prolic.comparison.SequenceBasedComparison;
import de.mpc.prolic.exceptions.EmptySequenceException;
import de.mpc.prolic.exceptions.InsufficientFileNumberException;
import de.mpc.prolic.exceptions.MissingFilenameExtensionException;
import de.mpc.prolic.exceptions.UnavailableProLiCFeatureException;
import de.mpc.prolic.exceptions.WrongProLiCSettingException;
import de.mpc.prolic.misc.CheckConfiguration;
import de.mpc.prolic.misc.IO;
import de.mpc.prolic.misc.Tools;
import de.mpc.prolic.model.*;

import java.util.ResourceBundle;
import java.util.concurrent.TimeUnit;


/**
 * @author Michael Kohl
 * The main class of ProLic. Manages the basic program sequence, i.e. this class represents the controller layer of the ProLiC software.
 */
public class Prolic {
	/*******************************************************************************
	 * Constants
	 ******************************************************************************/
	private final String paramAcc = "acc";										//The expected parameter for accession based comparison
	private final String paramSeq = "sequence";									//The expected parameter for sequence based comparison
	private final String paramPept = "peptide";									//The expected parameter for peptide based comparison
	private final String spacer = " ";

	/* The following constants are given in order to provide developers with an centralized overview of the setting parameters considered by the 
	command line version of ProLiC.	Here, the keys of the key - value pairs in the .ini file and the section names are given
	as constants (indicated by prefic k_).*/
	private final String nameConfFile = "ProLiC_conf.ini";						//Name of the configuration file
	
	// Logging
	private static final Logger log4j = LogManager.getLogger(Prolic.class 
	        .getName());
	// section genericSettings
	private final String sectionNameGenericSettings = "genericSettings";		//Name of the section in the configuration file with generic settings
																				//(i.e. required for both the command line and the planned GUI mode of ProLiC) 
	private final String k_colDelimIn = "column_delimiter_input";				//The specified column delimiter for the input file
	private final String k_colDelimOut = "column_delimiter_output";				//The specified column delimiter for the output file
	private final String k_fileType = "file_type";								//Specifies the type of information stored in the input file (protein information, gene information or both)
	private final String k_considerProteinNames="considerProteinNames";			//Specifies, whether protein name information is provided in the input files.
	private final String k_considerGeneNames="considerGeneNames";				//Specifies, whether gene name information is provided in the input files.
	private final String k_writePeptInfo="writePeptInfo";						//Specifies, whether peptide information is written in the result file.
	private final String k_encoding_output="encoding_output";					//Specifies the character set for writing ProLiC results to the output file
	//default values for generic settings (in order to prevent shutdown of ProLiC):
	private final String def_colDelimIn="\t";
	private final String def_colDelimOut="\t";
	private final String def_fileType="protein_information";
	private final String def_encoding_output="UTF-8";
	
	// section Enzymes
	private final String sectionNameEnzymeSettings = "Enzymes";					//Name of the section in the configuration file with settings that concern the used enzyme(s)
	// section commandLineSettings. These settings are read from the configuration file (a future GUI implementation is intended to support changing of these setting via the GUI)
	private final String sectionNameCLSettings = "commandLineSettings";			//Name of the section in the configuration file with settings required for running of ProLiC from the command line
	private final String k_inputDir = "inputDir";								//Absolute path of the input directory
	private final String k_outputDir = "outputDir";								//Absolute path of the output directory
	private final String k_fileExtIn = "file_extension_input";					//Specifies, which type of file should be read from the input directory
	private final String k_file_name_out = "file_name_out";						//The file name used for the output file that contains the results of the ProLiC comparison (extended with the timestamp)	
	private final String k_fileExtOut = "file_extension_output";				//Specifies, which type of file should be written in the output directory
	private final String k_compType="compType";									//Specifies the type of comparison (i.e. either accession, sequence or peptide based comparison)
	// Settings concerning modifications are necessary for both sequence and peptide based comparison
	private final String k_consid_mod="consideration_of_modifications";			//Specifies whether PTMs should be considered as distinctive features in sequence and peptide based comparisons.
	private final String k_regExModification="regExModification";				//Regular expression that is used to find a post translational modification in a one letter code protein sequence
	
	// section Acc. based comparison
	// private final String sectionNameAccBasedComp = "Acc. based comparison";	//Name of the section in the configuration file with settings related to an accession based comparison
	// Currently, there are no settings for an accession based comparison necessary  
	
	// section grouping
	private final String sectionNameGrouping = "grouping";						//Name of the section in the configuration file with settings related to grouping
	private final String k_selection_representative="selection_representative"; //The method utilized for calculation of a representative biomolecule for a biomolecule group. 
																				//Currently, allowed values are 'longestSequence' and 'random' only.
	// section Seq. based comparison
	private final String sectionNameSeqBasedComp = "Seq. based comparison"; 	//Name of the section in the configuration file with settings related to a sequnce based comparison
	private final String k_PID_method = "PID_method";							//Method for calculation of the percent sequence identity
	private final String k_PID_threshold = "PID_threshold";						//The percent sequence identity threshold, which needs to be passed in order to consider two sequences as identical 
	private final String k_subMatrix = "substitution_matrix";					//Name of the file that stores the substitution matrix 
	private final String k_subMatrixId = "substitution_matrix_id";				//Identifier for the substitution matrix (used for generation of the substitution matrix with BioJava)
	private final String k_alignMethod = "alignment_method";					//Name of the used alignment method (e.g. for different flavours of global sequence alignment)
	private final String k_gapOpen = "penalty_gap_open";						//Penalty for opening a gap
	private final String k_gapExt =  "penalty_gap_extension";					//Penalty for gap extension

	// section Peptide based comparison
	private final String sectionNamePeptBasedComp = "Peptide based comparison"; //Name of the section in the configuration file with settings related to a peptide based comparison
	private final String k_peptideBasedMethod="peptideBasedMethod";				//The used method for peptide based comparison

	/*Keys used within the result map of the pairwise sequence based comparison, corrrespond in SequenceBasedComparison.java**************************************/
	private final String k_pid="pid";											//Percent identity calculated according to the choosen PID method.
	/*******************************************************************************
	 * Members
	 ******************************************************************************/
	private String proLiC_modus;												//The current ProLiC usage modus (e.g. with GUI, as command line tool)  
	
	// Generic settings 
	private String colDelimIn;													//The specified column delimiter for the input file
	private String colDelimOut;													//The specified column delimiter for the output file
	private String fileType; 													//The information stored in the input file (protein information, gene information or both)
	private boolean considerProteinNames;										//Specifies, whether protein name information is provided in the input files and will be written to the output also
	private boolean considerGeneNames;											//Specifies, whether gene name information is provided in the input files and will be written to the output also
	private boolean writePeptInfo=false;										//Specifies, whether peptide information is written in the result file. 
	private String encoding_output;												//Specifies the character set for writing ProLiC results to the output file
	//Resource bundles
	ResourceBundle bundleMess = ResourceBundle.getBundle("de.mpc.prolic.resources.MessageBundle");		//A localized bundle for user information and for logging
	// Settings related to enzymes
	private ArrayList<Enzyme> enzymes;
	
	// Settings related to the ProLiC command line mode
	private String inputDir;													//Absolute path of the input directory
	private String outputDir;													//Absolute path of the output directory	
	private String fileExtIn;													//Specifies, which type of file should be read from the input directory
	private String file_name_out;												//The file name used for the output file that contains the results of the ProLiC comparison (extended with the timestamp)	
	private String fileExtOut;													//Specifies, which type of file should be written into output directory
	private String compType;													//The comparison method (accession, sequence or peptide based)
	private boolean consideration_of_modifications;								//Usage of PTMs for discrimination between protein list entries (yes/no)
	private String regExModification;											//Regular expression that is used to find a post translational modification in a one letter code protein sequence
	
	// Settings related to grouping
	private String selRepMethod="longestSequence";								//The method used for selection of group representatives.
																				//Default value is a random selection of biomolecule group representatives, 
																				//because this supports reproducibility of the results
	// Settings related to sequence based comparison
	private String PID_method;													//Method for calculation of the percent sequence identity 
	private double PID_threshold;												//The percent sequence identity threshold, which needs to be passed in order to consider two sequences as identical 
	//private String subMatrix;													//Name of the file that stores the substitution matrix (currently not in use)
	private String subMatrixId;													//Identifier for the substitution matrix (used for generation of the substitution matrix with BioJava)
	private String alignMethod;													//Name of the used alignment method (e.g. for different flavours of global sequence alignment)
	private int gapOpen=0;														//Penalty for opening a gap
	private int gapExt=0;														//Penalty for gap extension
	
	// Settings related to peptide based comparison
	private String peptideBasedMethod="includeSequenceInfo";					//The approach used for the calculation of an identity relationship between two compared proteins. Default is the method that takes protein sequence information into account.	
		
	// Data structures used during comparison
	private ArrayList<BiomoleculeList> inputLists;								//This lists stores lists generated directly from the files given in the input directory 								
	private ArrayList<Biomolecule> totalBiomoleculeList; 						//This list is generated from the inputLists and stores all the biomolecules that are contained within these lists
	private ArrayList <BiomoleculePair> relatedBiomolecules;					
	private Collection<Set<Biomolecule>> biomoleculeGroups;						//Collection of biomolecule groups, i.e. maximal cliques (each maximal clique is represented by a set of vertices)
	private TreeSet <BiomoleculeGroupList> compRes;								//This List is the final ProLiC result. compRes stores the sets of biomolecules (i.e. the intersections and
																				//the set differences, respectively that result from the comparison. compRes is also the used for writing the 
																				//result file. Here, TreeSet is an adequate collection because the list identifiers (=names) of the result lists 
																				//is unambiguous.
	// Miscellaneous
	private int maxNumberPeptides; 												//Stores the maximum number of peptides that are assigned to a protein that is considered for comparison.	
	private boolean debugMode=false;											//Several debugging tasks 
	private boolean debugMode2=false;											//Prints the current comparison number
	private Boolean inconsistenciesExist=false;									//Flag that indicates that there is occurrence of at least one unexpected event during the runtime of ProLiC. In this case logging and user information is given.


	/*******************************************************************************
	 * Constructors
	 ******************************************************************************/	
	/**
	 * This constructor is currently not in use. Intended for usage of ProLiC with a graphical user interface.
	 * ProLiC is then switched into the GUI mode.
	 */
	public Prolic(String GUI_) {
		proLiC_modus=GUI_;
		initGUI();
	}
	
	/**
	 * Empty constructor. Constructor intended for usage of ProLiC as a command line tool
	 */
	public Prolic() {
		checkProlicConfiguration();
		proLiC_modus="cl";
		initCL();		
	}	
	
	/**
	 * Main function. ProLiC expects no program arguments
	 * @param args
	 */
	public static void main(String[] args) {
		if (args.length==0){
			Prolic p = new Prolic();
			log4j.info(p.bundleMess.getString("m_program_start"));
			// the following construction is used for testing purposes
			try {
					p.readInputFromDir();				
		    } catch (MissingFilenameExtensionException mfee) {
		    	log4j.warn(mfee.getMessage());
		    	p.setInconsistenciesExist(true);
			}catch(InsufficientFileNumberException ife){
				IO.fatalErrorMessAndShutdown(ife.getMessage());
				if(p.getDebugMode()){
					System.out.println(ife.getStackTrace());
				}
			}catch(UnavailableProLiCFeatureException upf){
				IO.fatalErrorMessAndShutdown(upf.getMessage());
			}
			p.deleteDuplicates();
			p.buildTotalList();
			p.compare();
			p.compRes=p.prepareOutput(p.biomoleculeGroups);
			p.writeOutput();	
			if (p.getInconsistenciesExist()){
				log4j.warn(p.bundleMess.getString("m_warning_corrupted_results"));
				System.out.println(p.bundleMess.getString("m_warning_corrupted_results")+" "+p.bundleMess.getString("m_consider_log"));
				System.exit(2);
			}else{
				System.exit(0);
			}
		}
		if (args.length>0){
			System.out.println("There are too many program arguments. Please refer to the manual. Redundant parameters will be ignored.");
		}
	}
	
	/*******************************************************************************
	 * Initialization 
	 ******************************************************************************/		
	private void checkProlicConfiguration(){
		boolean success;
		String mess;
		Map<String, String> wv;
		Map<String, String> adv;
		Iterator <String> it;
		String currKey;
		CheckConfiguration cc = new CheckConfiguration(nameConfFile);
		success=cc.checkCompleteConfigurationFile();
		if(!success){
			mess=bundleMess.getString("m_wrong_parameter_values");
			System.out.println(mess);
			wv=cc.getWrongValues();
			it=wv.keySet().iterator();
			while(it.hasNext()){
				currKey=it.next();
				System.out.println(currKey+'='+wv.get(currKey));
			}			
			IO.fatalErrorMessAndShutdown(bundleMess.getString("m_program_shutdown"));
		}else{//checking whether or not default values must be applied.
			adv=cc.getAppliedDefaultValues();
			if(adv.size()>0){
				this.inconsistenciesExist=true;
				mess=bundleMess.getString("m_default_parameters_applied");
				System.out.println(mess);
				log4j.warn(mess);
				it=adv.keySet().iterator();
				while(it.hasNext()){
					currKey=it.next();
					System.out.println(currKey+'-'+adv.get(currKey));
					log4j.warn(currKey+'-'+adv.get(currKey));
				}
			}			
		}
	}
	
	
	/**
	 * Basic initialization of ProLiC when using the command line
	 */
	private void initCL(){
		Boolean success=true;
		Map <String, String> settings;
		String mess;		
		boolean doShutdown=true;//Reading of parameter settings from the .ini-file is required for a proper functioning of the command line version of ProLiC 
		//Basic initialization tasks for usage of ProLiC as a command line tool
		mess = bundleMess.getString("m_access_error_settings")+" "+sectionNameCLSettings+"."+spacer+bundleMess.getString("m_program_shutdown");
		settings=IO.readSectionFromIni(nameConfFile, sectionNameCLSettings, success, mess, doShutdown);
		if (success){
			inputDir=settings.get(this.k_inputDir);													
			outputDir=settings.get(this.k_outputDir);																										
			fileExtIn=settings.get(this.k_fileExtIn);
			file_name_out=settings.get(this.k_file_name_out);
			fileExtOut=settings.get(this.k_fileExtOut);		
			compType=settings.get(this.k_compType);
			consideration_of_modifications=Boolean.valueOf(settings.get(this.k_consid_mod));
			regExModification=settings.get(this.k_regExModification);
		}		
		initGeneric(); //Information about the value of compType is required for reading the 'generic' information from the configuration file
		//Shutdown in case of insufficient parameter values part 1
		if (inputDir==null||inputDir.equals("")||outputDir==null||outputDir.equals("")
				||fileExtIn==null||fileExtIn.equals("")||file_name_out==null||file_name_out.equals("")||
				fileExtOut==null||fileExtOut.equals("")||compType==null||compType.equals("")){
			mess = bundleMess.getString("m_access_error_settings2")+" "+bundleMess.getString("m_program_shutdown");
			IO.fatalErrorMessAndShutdown(mess);
		}
		//Checking whether the user given parameter is appropriate
		if (this.compType.equals(paramAcc)||compType.equals(paramSeq)||compType.equals(paramPept)){
			//do nothing
		} else {
			IO.fatalErrorMessAndShutdown(this.bundleMess.getString("m_compType_error")+" "+compType+"."+" "+this.bundleMess.getString("m_program_shutdown"));			
		}	

			
		//Initialization of ProLiC concerning the settings related to grouping. 
		mess = bundleMess.getString("m_access_error_settings")+" "+sectionNameGrouping+"."+spacer+bundleMess.getString("l_wrong_prolic_setting");
		settings=IO.readSectionFromIni(nameConfFile, this.sectionNameGrouping, success, mess, false); //no shutdown is necessary
		if(success){			
			selRepMethod=settings.get(this.k_selection_representative);		
		}else{
			inconsistenciesExist=true;
			log4j.warn(mess);
		}
		
		//Initialization of ProLiC concerning the settings related to sequence based comparison. 
		if (this.compType.equals(paramSeq)){
			mess = bundleMess.getString("m_access_error_settings")+" "+sectionNameSeqBasedComp+"."+" "+bundleMess.getString("m_program_shutdown"); 
			settings=IO.readSectionFromIni(nameConfFile, sectionNameSeqBasedComp, success, mess, doShutdown);
			if(success) {
				PID_method=settings.get(this.k_PID_method);	
				PID_threshold=Double.valueOf(settings.get(this.k_PID_threshold));
				subMatrixId=settings.get(this.k_subMatrixId);			
				alignMethod=settings.get(this.k_alignMethod);						
				gapOpen=Integer.valueOf(settings.get(this.k_gapOpen));														
				gapExt=Integer.valueOf(settings.get(this.k_gapExt));							
			}
			if (PID_method==null||PID_method.equals("")||subMatrixId==null||subMatrixId.equals("")|| 
					subMatrixId==null||subMatrixId.equals("")||alignMethod==null||alignMethod.equals("")){
				mess = bundleMess.getString("m_access_error_settings3")+"."+" "+bundleMess.getString("m_program_shutdown");
				IO.fatalErrorMessAndShutdown(mess);
			}
		}
		
		//Initialization of ProLiC concerning the settings related to peptide based comparison. 
		if(this.compType.equals(paramPept)){
			mess = bundleMess.getString("m_access_error_settings")+sectionNamePeptBasedComp+"."+" "+bundleMess.getString("m_program_shutdown"); 
			settings=IO.readSectionFromIni(nameConfFile, sectionNamePeptBasedComp, success, mess, doShutdown);
			if(success) {
				peptideBasedMethod=settings.get(this.k_peptideBasedMethod);
			}
			if (peptideBasedMethod==null||peptideBasedMethod.equals("")){ 
				mess = bundleMess.getString("m_access_error_settings4")+"."+" "+bundleMess.getString("m_program_shutdown");
				IO.fatalErrorMessAndShutdown(mess);
			}
		}
	}
	
	/**
	 * Basic initialization of ProLiC when using the graphical user interface (GUI)
	 */
	private void initGUI(){
		initGeneric();
	}
	
	/**
	 * Necessary initializations independent from the type of usage of ProLiC (e.g. as a command line tool or with the ProLiC GUI. In the
	 * latter case these settings are the default values). 
	 * Insufficient settings lead to program shutdown in case of ProLiC usage as a command line tool. 
	 */
	private void initGeneric(){
		boolean doShutdown=false;		
		Boolean success=true;
		String mess_generic = bundleMess.getString("m_access_error_settings")+sectionNameGenericSettings+"."+" "+bundleMess.getString("m_program_shutdown");
		String mess_enzymes = bundleMess.getString("m_access_error_settings")+sectionNameEnzymeSettings+"."+" "+bundleMess.getString("m_program_shutdown");
		Iterator<String>it;
		Enzyme aEnzyme=null;
		Map <String, String> settings;
		if (this.proLiC_modus.equals("cl")){
			doShutdown=true;
		}
		settings=IO.readSectionFromIni(nameConfFile, sectionNameGenericSettings, success, mess_generic, doShutdown);
		if (success) {		
			colDelimIn=settings.get(this.k_colDelimIn);													
			colDelimOut=settings.get(this.k_colDelimOut);
			fileType=settings.get(this.k_fileType);
			considerProteinNames=Boolean.valueOf(settings.get(this.k_considerProteinNames));
			considerGeneNames=Boolean.valueOf(settings.get(this.k_considerGeneNames));
			writePeptInfo=Boolean.valueOf(settings.get(k_writePeptInfo));
			encoding_output=settings.get(this.k_encoding_output);
		}
		if (colDelimIn==null||colDelimIn.equals("")){
			colDelimIn=this.def_colDelimIn;
		}
		if (colDelimOut==null||colDelimOut.equals("")){
			colDelimOut=this.def_colDelimOut;
		}
		if (fileType==null||fileType.equals("")){
			fileType=this.def_fileType;
		}
		if (encoding_output==null||encoding_output.equals("")){
			encoding_output=this.def_encoding_output;
		}
		
		doShutdown=true;
		//Initialization of ProLiC concerning information about the used enzyme.
		if(compType.equals(paramPept)){//accession or sequence based comparison requires no enzyme information
			doShutdown=true;
			settings=IO.readSectionFromIni(nameConfFile, sectionNameEnzymeSettings, success, mess_enzymes, doShutdown);
			if(success){
				enzymes=new ArrayList<Enzyme>();
				it=settings.values().iterator();
				while(it.hasNext()){			
					try{
						aEnzyme=Tools.createEnzyme(it.next());
					}catch(WrongProLiCSettingException wpse){
						log4j.warn(wpse.getMessage());
						inconsistenciesExist=true;
					}
					this.enzymes.add(aEnzyme); 
				}
			}
		}
		
		this.inputLists = new ArrayList<BiomoleculeList>();
		this.totalBiomoleculeList=new ArrayList<Biomolecule>();
		this.compRes=new TreeSet <BiomoleculeGroupList>();
	}
	
	/**
	 * The method checks different settings given by the configuration file and replaces incompatible settings with default values.
	 * Currently not in use. Created as an reminder for this opportunity.
	 */
	private void checkSettingCompatibility(){
				
	}
	
	/*******************************************************************************
	 * Preprocessing
	 ******************************************************************************/
	/**
	 * ProLiC does not accept duplicate entries (i.e. biomolecules with identical accession numbers) in the biomolecule lists. 
	 * This method removes duplicates within each list that was generated from the specified input files. 
	 */
	private void deleteDuplicates(){
		BiomoleculeList aBL;
		for (int i=0;i<inputLists.size();i++){
			aBL=calcNonRedundantList(inputLists.get(i));	
			inputLists.set(i, aBL);
		}
	}
	

	/**
	 * This method removes all duplicates from the BiomoleculeList (i.e. a list generated from a file in the input directory) specified in the argument of the function call.
	 * To this end, an instance of Map is used to remove the duplicates. This means that the last biomolecule with a distinct 
	 * accession number is maintained. All other biomolecules identified by this accession number, which were added to this map 
	 * previously, are dropped.
	 * @param aBL
	 * @return an instance of BiomoleculeList, which contains no biomolecule duplicates.
	 */
	private  BiomoleculeList calcNonRedundantList(BiomoleculeList aBL){
		String mess;
		HashMap<String,Biomolecule> noDups = new HashMap<String, Biomolecule>();
		TreeMap<String,Biomolecule> noDupsSorted = new TreeMap<String, Biomolecule>(
				new Comparator<String>() {					 
					@Override
					public int compare(String o1, String o2) {
						return o2.compareTo(o1);
					}
		 
				});
		Biomolecule aB;
		BiomoleculeList aBLnoDups = new BiomoleculeList();
		ArrayList<Biomolecule> biomoleculesNoDups=new ArrayList<Biomolecule>();
		ArrayList<Biomolecule> biomolecules = aBL.getBiomolecules(); 		 
		for (int i=0; i<biomolecules.size(); i++){
			aB = biomolecules.get(i);
			noDups.put(aB.getIdentifier(), aB);
		}
		noDupsSorted.putAll(noDups);
		biomoleculesNoDups.addAll(noDupsSorted.values());
		aBLnoDups.setBiomolecules(biomoleculesNoDups);
		aBLnoDups.setListName(aBL.getListName());
		if (aBL.getBiomolecules().size()>biomoleculesNoDups.size()){
			mess = aBL.getListName()+": "+bundleMess.getString("m_duplicates_removed");
			log4j.warn(mess);
			inconsistenciesExist=true;
		}
		return aBLnoDups;
	}
	
	/**
	 * Building a list that contains each biomolecule from the loaded input lists. Therefore this list contains the entirety of
	 * biomolecules that are considered for comparision
	 */
	private void buildTotalList(){
		BiomoleculeList aBL;
		for(int i=0; i<inputLists.size();i++){
			aBL=inputLists.get(i);
			totalBiomoleculeList.addAll(aBL.getBiomolecules());				
		}
	}
	
	/**
	 * Enables removing of modifications in both the protein sequences all well in the peptides of all proteins, which are stored in the
	 * totalBiomoleculeList. Genes that may also part of this list are ignored.
	 */
	private void removeAllModifications(){
		Biomolecule aBM;
		Protein aProt;
		Peptide aPept;
		for (int i=0; i<totalBiomoleculeList.size();i++){
			aBM=totalBiomoleculeList.get(i);
			if(aBM.getBiomoleculeType().equalsIgnoreCase("protein")){
				try{
					aProt=(Protein)aBM;
					aProt.setSequence(Tools.removePTMs(aProt.getSequence(), this.regExModification));
					for (int j=0; j<aProt.getPeptides().size();j++){
						aPept=aProt.getPeptides().get(j);
						aPept.setSequence(Tools.removePTMs(aPept.getSequence(), this.regExModification));
					}			
				}catch(WrongProLiCSettingException wpse){
					log4j.warn(bundleMess.getStringArray("l_wrong_prolic_setting")+this.spacer+regExModification+".");
					this.inconsistenciesExist=true;
				}
			}
		}		
	}
	
	/*******************************************************************************
	 * Methods related to comparison, the subsequent grouping and for sorting the data structure
	 ******************************************************************************/		
	
	/**
	 * This method directs the comparison, i.e. different methods are invoked depending on the specified comparison type.
	 */
	private void compare(){
		long limitMiliSeconds=100;		//Specifies the limit for presinting the output of the comparison time in milliseconds.
		long limitSeconds=100000;		//Specifies the limit for calculation of milliseconds into seconds. Corresponds to 100 seconds.
		long limitMinutes=6000000;		//Specifies the limit for calculation of milliseconds into minutes. Corresponds to 100 minutes.
		long limitHours=360000000;		//Specifies the limit for calculation of milliseconds into hours. Corresponds to 100 hours.
		long comparisonTime=0;			//The time needed for the comparison
		long comparisonTimeStart=0;		//The system time at the time where the comparison starts.
		long comparisonTimeEnd=0;		//The system time at the time where the comparison ends.
		//Calendar c = new GregorianCalendar();
		String timeUnit="";
		if (this.proLiC_modus.equals("cl")){
			System.out.println(this.bundleMess.getString("m_comparison_start")+" "+Tools.getCurrentDate()+".");
			log4j.info(this.bundleMess.getString("m_comparison_start")+" "+Tools.getCurrentDate()+".");
		}
		relatedBiomolecules=new ArrayList <BiomoleculePair>();
		if (compType.equals(paramAcc)){ //Accession based comparison and in this case also the grouping
			System.out.println(this.bundleMess.getString("m_comparison_acc"));
			log4j.info(this.bundleMess.getString("m_comparison_acc"));
			comparisonTimeStart=System.currentTimeMillis();
			compAcc();
			BuildCliquesSG();
			comparisonTimeEnd=System.currentTimeMillis();
		}
		//Performing the pairwise comparison depending on the specified comparison type (either sequence or peptide based comparison)		
		if (compType.equals(paramSeq)){ //Sequence based comparison
			System.out.println(this.bundleMess.getString("m_comparison_seq"));
			log4j.info(this.bundleMess.getString("m_comparison_seq"));
			comparisonTimeStart=System.currentTimeMillis();
			removeAllModifications();
			compSeq();
			BuildCliquesSG();
			comparisonTimeEnd=System.currentTimeMillis();
		}
		if (compType.equals(paramPept)){ //Peptide based comparison
			System.out.println(this.bundleMess.getString("m_comparison_pept"));
			log4j.info(this.bundleMess.getString("m_comparison_pept"));			
			comparisonTimeStart=System.currentTimeMillis();
			compPept();
			BuildCliquesSG();
			comparisonTimeEnd=System.currentTimeMillis();
		}
		if (this.proLiC_modus.equals("cl")){
			System.out.println(this.bundleMess.getString("m_comparison_end")+" "+Tools.getCurrentDate()+".");
			log4j.info(this.bundleMess.getString("m_comparison_end")+" "+Tools.getCurrentDate()+".");
		}
		comparisonTime=comparisonTimeEnd-comparisonTimeStart;
		if (comparisonTime<=limitMiliSeconds){
			timeUnit="ms";
		}
		if (comparisonTime>limitMiliSeconds && comparisonTime<=limitSeconds){
			comparisonTime = TimeUnit.MILLISECONDS.toSeconds(comparisonTime);
			timeUnit="s";
		}
		if (comparisonTime>limitSeconds && comparisonTime<=limitMinutes){
			comparisonTime = TimeUnit.MILLISECONDS.toMinutes(comparisonTime);
			timeUnit="min";
		}
		if (comparisonTime>limitMinutes && comparisonTime<=limitHours){
			comparisonTime = TimeUnit.MILLISECONDS.toHours(comparisonTime);
			timeUnit="h";
		}
		if (comparisonTime>limitHours){
			comparisonTime = TimeUnit.MILLISECONDS.toDays(comparisonTime);
			timeUnit="d";
		}
		if (this.proLiC_modus.equals("cl")){
			System.out.println(this.bundleMess.getString("m_comparison_time")+" "+comparisonTime+" "+timeUnit+".");
			log4j.info(this.bundleMess.getString("m_comparison_time")+" "+comparisonTime+" "+timeUnit+".");
		}		
	}
	
	
	/**
	 * Method calculates maximal cliques (=Biomolecule groups) according to the bilateral similarity relationship of two proteins in an undirected graph.
	 * Therefore, the Bron-Kerbosch algorithm is used. Uses the class SimpleGraph of the JGraphT library.
	 */
	private void BuildCliquesSG(){
	    UndirectedGraph<Biomolecule, DefaultEdge>completeGraph=
	        new SimpleGraph<Biomolecule, DefaultEdge>(DefaultEdge.class);
	// add the vertices. Because possible duplicates in the input lists are removed the relatedProtList entries correspond to the vertices
	    for (int i=0; i<totalBiomoleculeList.size();i++){
	    	completeGraph.addVertex(totalBiomoleculeList.get(i));
	    }
	// add edges to create a relationship in the graph, each relationship in the relatedProtList corresponds to an edge
	    for (int j=0; j<this.relatedBiomolecules.size();j++){
	    	BiomoleculePair aBP = relatedBiomolecules.get(j);
	    	//always two entries in the vector (pairwise related proteins)
	    	Biomolecule bm1=aBP.getBiomoleculePair().get(0);
	    	Biomolecule bm2=aBP.getBiomoleculePair().get(1);
	    	completeGraph.addEdge(bm1, bm2);
	    }
	// retrieving the maximal cliques
		BronKerboschCliqueFinder <Biomolecule, DefaultEdge> CF=new BronKerboschCliqueFinder<Biomolecule, DefaultEdge>(completeGraph);
		this.biomoleculeGroups= CF.getAllMaximalCliques();
		if (debugMode){
			writeBiomoleculeGroups();
		}
	}
	
	/**
	 * This method calculates different lists (particularly the intersections and set differences) starting with the calculated maximal cliques.
	 * Additionally, different alphabetic sorting procedures are carried out: 
	 * First, the biomolecules within a given biomolecule group are sorted according to their unique identifiers. 
	 * Second, the biomolecule groups within a list are sorted according to the unique identifier of the specified biomolecule group representatives and 
	 * third the resulting comparison lists are sorted with respect of the list names.
	 */
	private TreeSet <BiomoleculeGroupList> prepareOutput(Collection<Set<Biomolecule>> biomoleculeGroups_){
		int currNumbBg=0;
		Random rand = new Random();
		boolean listExists=false;
		TreeSet <BiomoleculeGroupList> res = new TreeSet <BiomoleculeGroupList>();
		BiomoleculeGroupList aBGL;
		BiomoleculeGroup aBG;
		Iterator <Set<Biomolecule>> iter = biomoleculeGroups_.iterator();
		Iterator <BiomoleculeGroupList> iter2;
		String bgListIdentifier;
		while (iter.hasNext()){
			listExists=false;
			Set<Biomolecule> aBMGroup = iter.next();
		//Creating an instance of BiomoleculeGroup
			aBG=new BiomoleculeGroup();
			aBG.addBiomolecules(aBMGroup);
			aBG.setId(new Integer(currNumbBg));
			currNumbBg++;
			aBG.setRepresentative(this.selectRepresentative(aBG, this.selRepMethod,rand.nextInt(aBG.getBiomoleculeGroup().size())));
		//Building a list identifier for this BiomoleculeGroup	
			bgListIdentifier=Tools.buildListIdentifier(aBG);
		//Adding the current BiomoleculeGroup to an existing BiomoleculeGroupList or otherwise creating a new BiomoleculeGroupList
			iter2=res.iterator();
			while (iter2.hasNext()){
				aBGL=iter2.next();
				if (aBGL.getListIdentifier().equals(bgListIdentifier)){
					aBGL.addBiomoleculeGroup(aBG);
					listExists=true;
				}
			}
			if (!listExists){
				aBGL=new BiomoleculeGroupList();
				aBGL.addBiomoleculeGroup(aBG);
				aBGL.setListIdentifier(bgListIdentifier);
				res.add(aBGL);
			}
		}
		//Sorting of the biomolecule groups within a biomolecule group list
		iter2=res.iterator();
		while (iter2.hasNext()){
			iter2.next().sortBiomoleculeGroups();
		}
		return res;
	}
	
	/**
	 * This method calculates the representative for a given BiomoleculeGroup
	 * @param aBG
	 * @param selMethod
	 * @param selIndex
	 * @return an instance of Biomolecule that is selected as representative for this biomolecule group.
	 */
	private Biomolecule selectRepresentative(BiomoleculeGroup aBG, String selMethod, int selIndex){
		Biomolecule aBm=null;
		Biomolecule selBm;
		String aSeq;
		ArrayList<Biomolecule> BmArrList=new ArrayList<Biomolecule>();
		BmArrList.addAll(aBG.getBiomoleculeGroup());
		if (selMethod.equals("longestSequence")){
			if(checkExistenceOfSequence(aBG)){
				selBm=BmArrList.get(0);
				for (int i=1;i<BmArrList.size();i++){
					aBm=BmArrList.get(i);
					aSeq=aBm.getSequence();
					if (aSeq.length()>selBm.getSequence().length()){
						selBm=aBm;
					}
				}
			}else{//perform random selection of the group representative
				selBm=BmArrList.get(selIndex);
				inconsistenciesExist=true;
				log4j.warn(bundleMess.getString("l_representative_selection_error2")+" "+bundleMess.getString("l_failed_grouping_settings"));
			}	
		}else{
			if (selMethod.equals("random")){
				selBm=BmArrList.get(selIndex);
			}else{//default method: random selection of the representative
				selBm=BmArrList.get(selIndex);
				inconsistenciesExist=true;
				log4j.warn(bundleMess.getString("l_representative_selection_error1")+" "+bundleMess.getString("l_failed_grouping_settings"));
			}			
		}
		return selBm;		
	}
	
	/**
	 * This method checks the existence of sequences for all biomolecules of this biomolecule group
	 * @param aBG
	 * @return a boolean indicating if all biomolecules of this biomolecule group contain sequences.
	 */
	private boolean checkExistenceOfSequence(BiomoleculeGroup aBG){
		boolean checkPassed=true;
		String seq;
		for (int i=0; i<aBG.getBiomoleculeGroup().size();i++){
			seq=aBG.getBiomoleculeGroup().get(i).getSequence();
			if((seq==null)||(seq.equals(""))){
				checkPassed=false;
			}
		}
		return checkPassed;
	}
	/*******************************************************************************
	 * Accession based comparison
	 ******************************************************************************/	
	/**
	 * The compAcc method performs an accession based comparison of biomolecule identifiers (e.g. 
	 * comparison of protein accession numbers) and the construction of the biomolecule groups
	 */
	private void compAcc(){
		Biomolecule aB;
		Biomolecule aBToComp;
		BiomoleculePair aBP;
		for (int i=0; i<totalBiomoleculeList.size();i++){
			aB = this.totalBiomoleculeList.get(i);
			for (int j=i+1; j<totalBiomoleculeList.size();j++){
				aBToComp = totalBiomoleculeList.get(j);
				if (aB.getIdentifier().equals(aBToComp.getIdentifier())){
					aBP = new BiomoleculePair();
					aBP.addBiomolecule(aB);
					aBP.addBiomolecule(aBToComp);
					this.relatedBiomolecules.add(aBP);
				}
			}
		}	
	}
	
	/*******************************************************************************
	 * Sequence based comparison
	 ******************************************************************************/
	
	/**
 	 * The compSeq method performs an sequence based comparison of two biomolecule sequences (e.g. the comparison of protein sequences)
	 */
	private void compSeq(){
		int k=0;
		HashMap <String, String> hmResults;
		double pid;
		String mess;
		SequenceBasedComparison sbc = new SequenceBasedComparison();
		sbc.setSubstMatrixID(this.subMatrixId);
		sbc.setPidMethod(PID_method);
		sbc.setGapOpen(gapOpen);
		sbc.setGapExt(gapExt);
		Biomolecule aBM;
		Biomolecule aBMtoComp;
		BiomoleculePair aBP;
		if(debugMode2){//Required for performance checks
			System.out.println("Total number of biomolecules: "+totalBiomoleculeList.size());
		}
		for (int i=0;i<this.totalBiomoleculeList.size();i++){
			aBM=totalBiomoleculeList.get(i);
			for (int j=i+1;j<totalBiomoleculeList.size();j++){
				aBMtoComp=totalBiomoleculeList.get(j);
				try {
					if(debugMode2){//Required for performance checks
						System.out.println("Current comparison: "+k);
						k++;
					}
					hmResults=sbc.pairwiseSequenceComp(aBM.getSequence(), aBMtoComp.getSequence());	
					pid=Double.valueOf(hmResults.get(this.k_pid));
					if (pid>=this.PID_threshold){ //true = the comparison passes the sequence identity criteria
	//Creating a new data structure for the relationship between two identifiedProteins
						aBP = new BiomoleculePair();
						aBP.addBiomolecule(aBM);
						aBP.addBiomolecule(aBMtoComp);
						this.relatedBiomolecules.add(aBP);	
					}
				} catch (EmptySequenceException esEx) {
					inconsistenciesExist=true;
					mess = this.bundleMess.getString("m_comparison_seq_empty_Sequence_error");
					log4j.error(mess);
				} catch (CompoundNotFoundException cnfEx) {
					inconsistenciesExist=true;
					log4j.error(cnfEx.getMessage());
				}
			}
		}	
	}
	
	/*******************************************************************************
	 * Peptide based comparison
	 ******************************************************************************/
	/**
 	 * The compPept method performs comparison of proteins based on peptide information. Different approaches for peptide based comparison are
 	 * supported with respect to the settings in the configuration file.
	 */
	private void compPept(){
		if((peptideBasedMethod.equals("includeSequenceInfo"))||(peptideBasedMethod.equals("peptidesOnly"))){
			compPept_ProtSeqOrPeptOnly();
		}	
		//TODO: Future implementation of further peptide based comparison methods
	}
	
	/**
	 * Peptide based comparison considering available protein sequence information or alternatively a comparison that refers to the identified peptide sets
	 * only. The algorithm assumes that all considered biomolecules are proteins.
	 */
	private void compPept_ProtSeqOrPeptOnly(){
		boolean isSimilar=false; 		//Stores the information whether two compared IdentifiedProteins are similar or not (with respect to given peptide information)
		Protein aProt;
		Protein anOtherProt;
		Protein aProtNoMod=new Protein();
		Protein anOtherProtNoMod=new Protein();
		BiomoleculePair aBP;
		String totBML_number="Total number of biomolecules in the considered lists: "+String.valueOf(totalBiomoleculeList.size());
		System.out.println(totBML_number);
		for (int i=0; i<totalBiomoleculeList.size();i++){
		//String mess_progress = "Number of executed comparisons: "+String.valueOf(i);
		//System.out.println(mess_progress);
			if((i%1000==0)){
				if (this.debugMode2){
					String mess_progress = "Number of currently executed comparisons (printed only for each thousandth comparison) : "+String.valueOf(i);
					System.out.println(mess_progress);
				}
			}
			aProt = (Protein)totalBiomoleculeList.get(i);
			for (int j=i+1; j<this.totalBiomoleculeList.size();j++){
				isSimilar=false;
				anOtherProt = (Protein)totalBiomoleculeList.get(j);
				if(peptideBasedMethod.equals("includeSequenceInfo")){
					if(!this.consideration_of_modifications){
						//Removing the the modification information from the protein sequences that should be compared
						aProtNoMod.setSequence(Tools.removePTMs(aProt.getSequence(), this.regExModification));						
						anOtherProtNoMod.setSequence(Tools.removePTMs(anOtherProt.getSequence(), this.regExModification));
						//Removing the the modification information from peptides of the proteins that should be compared
						aProtNoMod.setPeptides(Tools.removeModificationsFromPeptideSet(aProt.getPeptides(), this.regExModification));
						anOtherProtNoMod.setPeptides(Tools.removeModificationsFromPeptideSet(anOtherProt.getPeptides(), this.regExModification));
						isSimilar=PeptideBasedComparison.bilateralCompPeptSeq(aProtNoMod, anOtherProtNoMod, enzymes);
					}else{
						isSimilar=PeptideBasedComparison.bilateralCompPeptSeqMod(aProt, anOtherProt, enzymes);
					}					
				}
				if(peptideBasedMethod.equals("peptidesOnly")){
					if(!this.consideration_of_modifications){
						//Removing the the modification information from peptides of the proteins that should be compared
						aProtNoMod.setPeptides(Tools.removeModificationsFromPeptideSet(aProt.getPeptides(), this.regExModification));
						anOtherProtNoMod.setPeptides(Tools.removeModificationsFromPeptideSet(anOtherProt.getPeptides(), this.regExModification));
						isSimilar=PeptideBasedComparison.bilateralCompPeptOnly(aProtNoMod, anOtherProtNoMod);
					}else{
						isSimilar=PeptideBasedComparison.bilateralCompPeptOnly(aProt, anOtherProt);
					}					
				}				
				if(isSimilar){
					aBP = new BiomoleculePair();
					aBP.addBiomolecule(aProt);
					aBP.addBiomolecule(anOtherProt);
					this.relatedBiomolecules.add(aBP);				
				}
			}	
		}
	}
	
	/*******************************************************************************
	 * IO
	 ******************************************************************************/
	/**
	 * readInputFromDir reads biomolecule list information stored in an arbitrary number
	 * of tab separated (default) input files given in the input directory specified in the settings
	 * file. However, sub directories are not considered. Only files with the file extension specified in the configuration
	 * file are considered for further processing with ProLiC. The name of existing files with other extensions is logged in the log-file. 
	 * @throws MissingFilenameExtensionException 
	 */
	private void readInputFromDir() throws UnavailableProLiCFeatureException, MissingFilenameExtensionException, InsufficientFileNumberException{ 
		BiomoleculeList aBL=null;
		int maxNumbPeptides=0;
		int currMaxNumbPeptides;
		String[] filesList; //The list of file names in the specified input directory.
		String trimedCurrFileName;
		File directory = new File(inputDir);
		File currFile;
		filesList = directory.list();
		for (int i=0; i<filesList.length;i++){
				trimedCurrFileName = null;
				currFile=new File(directory,filesList[i]);
				if(currFile.isFile()){//considers only files but no directories or hidden files
					if(Tools.getFileExtension(filesList[i]).equalsIgnoreCase(this.fileExtIn)){//considers only files with the file extension specified in the configuration file
						trimedCurrFileName=Tools.trimFileExtension(filesList[i]);
						if (trimedCurrFileName!=null){
							if (fileType.equals("protein_information")){							
								aBL=IO.readProLiCformatedProteinFile(currFile, this.colDelimIn, trimedCurrFileName, 
										considerProteinNames, considerGeneNames, log4j, this.inconsistenciesExist);
								currMaxNumbPeptides=Tools.calcMaxNumberPeptids(aBL);
								if (currMaxNumbPeptides>maxNumbPeptides){
									maxNumbPeptides=currMaxNumbPeptides;
								}
							}
							if (fileType.equals("gene_information")){
								throw new UnavailableProLiCFeatureException(bundleMess.getString("m_upf_fileType_exception")+spacer+bundleMess.getString("m_program_shutdown"));
							}
							if (fileType.equals("mixed")){
								throw new UnavailableProLiCFeatureException(bundleMess.getString("m_upf_fileType_exception")+spacer+bundleMess.getString("m_program_shutdown"));
							}
							inputLists.add(aBL);
						}else{
							throw new MissingFilenameExtensionException(bundleMess.getString("l_missing_input_file_ext")+spacer+currFile+".");
						}
					}else{/* in case of files in the input directory that comprise a file extension that does not match the value of the file_extension_input key in the configuration file
						an entry will be written in the configuration file.*/
			    		if(debugMode){
			    			System.out.println(bundleMess.getString("l_wrong_input_file_ext")+spacer+currFile+".");
			    		}
			    		log4j.warn(bundleMess.getString("l_wrong_input_file_ext")+spacer+currFile+".");
			    		this.inconsistenciesExist=true;
						//new MalformedFilenameException(bundleMess.getString("l_wrong_input_file_ext")+spacer+currFile+".");
					}
				}
		} 
		this.maxNumberPeptides=maxNumbPeptides;
		if (inputLists.size()<2){ // a minimum of two files (= experiments) is required 
			throw new InsufficientFileNumberException(bundleMess.getString("m_insufficient_file_number")+spacer+bundleMess.getString("m_program_shutdown"));
		} 
	}
	
	/**
	 * writeOutput writes the results of the comparison to a tab separated (default)
	 * in the specified output directory (c.f. configuration file)
	 */
	private void writeOutput(){
		IO.writeResults(this.compRes, file_name_out, fileExtOut, outputDir, 
				colDelimOut, considerProteinNames, considerGeneNames, writePeptInfo, maxNumberPeptides, log4j, this.inconsistenciesExist, this.encoding_output);
		String mess;
		mess = this.bundleMess.getString("m_program_shutdown3");
		System.out.println(mess);
		System.out.println("exit(0)");
		log4j.info(mess);
		log4j.info("exit(0)");
	}	
	

	/*******************************************************************************
	 * Debugging
	 ******************************************************************************/	
	/**
	 * For debugging. Writes the biomolecule groups calculated as maximal cliques to the console
	 */
	private void writeBiomoleculeGroups(){
		Biomolecule aBm;
		Iterator <Set<Biomolecule>> it=biomoleculeGroups.iterator();
		Set<Biomolecule> aSet;
		Iterator <Biomolecule> it2;
        it=biomoleculeGroups.iterator();	
		  while (it.hasNext()){
			  aSet=it.next();
			  if(this.debugMode){
				  //System.out.println("Biomoleklgruppe:");
			  }			  
			  it2=aSet.iterator();
			  while (it2.hasNext()){
				  aBm=it2.next();
				  if(this.debugMode){
					  //System.out.println(aBm.getIdentifier()+";"+aBm.getOrigin());
				  }				  
			  }
		  }
	}
	
	/*******************************************************************************
	 * Methods for testing
	 * Hint: all methods are private in order to maintain encapsulation. They are
	 * invoked by usage of the methods specified in the class ProlicProxy.java.
	 ******************************************************************************/
	
	/**
	 * This setter is used for unit testing.
	 */
	private void setTotalBiomoleculeList(ArrayList<Biomolecule> tBMList){
		this.totalBiomoleculeList=tBMList;
	}
	
	/**
	 * This setter is used for unit testing.
	 */
	private void setRelatedBiomolecules(ArrayList <BiomoleculePair> rBM){
		this.relatedBiomolecules=rBM;
	}
	
	/**
	 * This setter is used for unit testing.
	 */
	private void setInputLists(ArrayList<BiomoleculeList> inputLists_){
		this.inputLists=inputLists_;
	}
	
	/**
	 * This getter is used for unit testing.
	 */
	private ArrayList<BiomoleculeList> getInputLists(){
		return this.inputLists;
	}
	
	/**
	 * This setter is used for unit testing.
	 * @param inconsistenciesExist_
	 */
	private void setInconsistenciesExist(boolean inconsistenciesExist_){
		inconsistenciesExist=inconsistenciesExist_;
	}
	
	/**
	 * This setter is used for unit testing.
	 * @param selRepMethod_
	 */
	private void setRepresentativeMethod(String selRepMethod_){
		selRepMethod=selRepMethod_;
	}

	/*******************************************************************************
	 * Setter and Getter
	 ******************************************************************************/	
	
	public boolean getInconsistenciesExist(){
		return inconsistenciesExist;
	}
	

	public boolean getDebugMode(){
		return this.debugMode;
	}	
}
